import contextlib
from typing import Optional, List

from prefect_streamline.core import deploy as core_deploy, importutils

_deploy_book = core_deploy.DeployBook()


def deploy() -> None:
    core_deploy.deploy(_deploy_book)


def list() -> List[core_deploy.DeployFlow]:
    return core_deploy.list(_deploy_book)


def register(name: Optional[str] = None, interval: Optional[int] = None, cron: Optional[str] = None):
    """
    register a flow to deploy it with the command prefect-streamline deploy

    >>> from prefect import flow
    >>> from prefect_streamline import deploybook
    >>>
    >>> @deploybook.register(interval=60)
    >>> @flow(name="main.my_favorite_function")
    >>> def myflow() -> int:
    >>>     return 43

    it is possible to save several deployments for the same flow.

    >>> from prefect import flow
    >>> from prefect_streamline import deploybook
    >>>
    >>> @deploybook.register(interval=60)
    >>> @deploybook.register(interval=3600, name="daily")
    >>> @flow(name="main.myflow")
    >>> def myflow() -> int:
    >>>     return 43

    it is possible to use a cron expression instead of an interval but not both at the same time in the same record.

    >>> from prefect import flow
    >>> from prefect_streamline import deploybook
    >>>
    >>> @deploybook.register(interval=60)
    >>> @deploybook.register(cron="0 0 * * *", name="daily")
    >>> @flow(name="main.myflow")
    >>> def myflow() -> int:
    >>>     return 43
    """

    def inner(func):
        core_deploy.register(_deploy_book, func, name=name, interval=interval, cron=cron)
        return func

    return inner


@contextlib.contextmanager
def use_dedicated_deploybook():
    global _deploy_book
    current_deploy_book = _deploy_book
    _deploy_book = core_deploy.DeployBook()
    yield
    _deploy_book = current_deploy_book


def import_module(module: str) -> None:
    importutils.import_module(module)


def import_path(path: str, discover: bool = False) -> None:
    importutils.import_path(path, discover)



"""
This module provides a library to plan a deployment for Prefect.
The developer saves these flows in a deployment book.

>>> deploy_book = deploy.create_deploy_book()
>>> deploy.register(deploy_book, flow1)
>>> deploy.register(deploy_book, flow2)
>>> deploy.deploy(deploy_book)
"""
import dataclasses
import os
from typing import List, Optional, Union, cast

from prefect import Flow
from prefect.deployments import Deployment
try:
    from prefect.server.schemas.schedules import IntervalSchedule, CronSchedule
except ImportError as exception:
    from prefect.orion.schemas.schedules import IntervalSchedule, CronSchedule # type: ignore



@dataclasses.dataclass
class DeployFlow:
    """
    defines a flow to deploy
    """
    flow: Flow
    name: Optional[str] = None
    interval: Optional[int] = None
    cron: Optional[str] = None

    def __post_init__(self):
        if self.name is None:
            self.name = "main"

    @property
    def schedule(self) -> Optional[Union[IntervalSchedule, CronSchedule]]:
        if self.interval is None and self.cron is None:
            return None
        elif self.interval is not None:
            return IntervalSchedule(interval=self.interval)
        elif self.cron is not None:
            return CronSchedule(cron=self.cron)

        return None

    def format(self):
        label = f"{self.flow.__name__}" + " ("
        elements = []
        if self.name is not None:
            elements.append(f"name:{self.name}")
        if self.interval is not None:
            elements.append(f"interval:{self.interval}")
        if self.cron is not None:
            elements.append(f"cron:{self.cron}")

        return label + ", ".join(elements) + ")"

@dataclasses.dataclass
class DeployBook:
    deploy_flows: List[DeployFlow] = dataclasses.field(default_factory=list)


def create_deploy_book() -> DeployBook:
    return DeployBook()


def register(deploy_book: DeployBook, flow: Flow, name: Optional[str] = None, interval: Optional[int] = None, cron: Optional[str] = None) -> None:
    if interval is not None and cron is not None:
        raise ValueError(f"Cannot define both interval and cron for the flow {flow.name=} - {interval=} - {cron=}")

    deploy_book.deploy_flows.append(DeployFlow(flow=flow, name=name, interval=interval, cron=cron))


def deploy(book: DeployBook) -> None:
    """
    The url of the prefect instance to deploy to is retrieved
    from the PREFECT_API_URL environment variable

    Inject a specific PREFECT_API_URL variable of the endpoint the deploy script will use.
    """
    root_dir = os.path.join(os.path.dirname(__file__), '..', '..', '..')
    for deploy_flow in book.deploy_flows:
        deployment = Deployment.build_from_flow(
            flow=deploy_flow.flow,
            name=cast(str, deploy_flow.name),
            path=root_dir,
            schedule=deploy_flow.schedule,
        )

        deployment.apply() # type: ignore


def list(book: DeployBook):
    return book.deploy_flows



import importlib
import importlib.util
import os
from os.path import basename
from types import ModuleType
from typing import List, Optional


def import_module(module: str) -> None:
    importlib.import_module(module)
    return


def import_path(path: str, discover: bool = False) -> None:
    if discover is True:
        _import_modules_from_path(path, recursive=True)
    else:
        module_name = basename(path).split(".")[0]
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is not None and spec.loader is not None:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        return


def _import_modules_from_path(path: str,
                              module_fullname=None,
                              ignore_files: Optional[List[str]] = None,
                              recursive=False) -> List[ModuleType]:
    if ignore_files is None:
        ignore_files = []

    full_path = os.path.realpath(path)
    assert os.path.isdir(full_path), f"{full_path} is not a directory"

    python_files = [f for f in os.listdir(full_path)
                    if os.path.isfile(os.path.join(full_path, f)) and f.endswith(".py")]

    modules = []
    for python_file in python_files:
        if python_file not in ignore_files:
            python_module = python_file.replace('.py', '')

            package_def = [] if module_fullname is None else [module_fullname]
            package_def.append(python_module)
            python_module_full_name = ".".join(package_def)

            spec = importlib.util.spec_from_file_location(python_module_full_name, os.path.join(full_path, python_file))
            if spec is not None and spec.loader is not None:
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                modules.append(module)

    if recursive:
        packages = [f for f in os.listdir(full_path) if os.path.isdir(os.path.join(full_path, f))]
        for package in packages:
            modules += _import_modules_from_path(os.path.join(path, package),
                                                 f"{module_fullname}.{package}",
                                                 ignore_files=ignore_files,
                                                 recursive=recursive)

    return modules



from typing import Optional

import click
from click import UsageError

from prefect_streamline import deploybook

@click.group()
def cli():
    pass


@cli.command('deploy')
@click.option('--discover', '-d', is_flag=True, help='discover flows in all submodules')
@click.option('--list', '-l', '_list', is_flag=True, help='list all the deployments', metavar='list_')
@click.option('--module', '-m', help='module to look flows in')
@click.argument('path', nargs=1, required=False)
def deploy_cmd(discover: bool, _list: bool, module: Optional[str] = None, path: Optional[str] = None):
    """
    deploy will deploy flow to prefect register with deploybook.register.

    \b
    $ prefect_streamline deploy src/app/flow.py
    $ prefect_streamline deploy -d src/app

    deploy can use module instead of path.

    \b
    $ prefect_streamline deploy -m app.flow
    """
    from prefect_streamline.core import deploy

    if module is None and path is None:
        raise UsageError(f"either module or or path must be defined")

    if module is not None and path is not None:
        raise UsageError(f"either module or or path must be defined, not both")

    if module is not None and discover is True:
        _echo_warning("option --discover/-d is not compatible with option --module/-m")

    if module is not None:
        deploybook.import_module(module)

    if path is not None:
        deploybook.import_path(path, discover=discover)

    if _list is False:
        deploybook.deploy()
    else:
        deploys = deploybook.list()
        for _deploy in deploys:
            click.echo(_deploy.format())


def _echo_warning(text: str) -> None:
    click.echo(click.style(f"Warning: {text}", fg='yellow'))
